/*
 * Copyright (c) 1996-1999 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */

// tclTunable.C
// C++ code that provides access to tunable constants from tcl.

/* $Id: tclTunable.C,v 1.20 2001/06/20 20:38:34 schendel Exp $ */

#include <assert.h>
#include <stdlib.h> // atoi()

#include "common/h/headers.h"
#include "tcl.h"
#include "tk.h"
#include "../TCthread/tunableConst.h"

#include "tclTunable.h"
#include "pdutil/h/TclTools.h"

struct cmdTabEntry {
   const char *cmdName;
   int   index;
   int   numArgs;
};

const int GETBOOLALLNAMES     =0;
   // return an array of names, representing the associative
   // array index values for each boolean tunable constant in
   // the registry.
const int GETFLOATALLNAMES    =1;
   // return an array of names, representing the associative
   // array index values for each float tunable constant in
   // the registry.
const int GETNUMTUNABLES      =2;
const int GETNUMBOOLTUNABLES  =3;
const int GETNUMFLOATTUNABLES =4;
const int GETDESCRIPTION      =5;
const int GETVALUEBYNAME      =6;
const int SETVALUEBYNAME      =7;
const int GETTYPEBYNAME       =8;
const int GETUSEBYNAME        =9;
const int GETFLOATRANGEBYNAME =10;
const int CMDERROR            =11;

struct cmdTabEntry TclTunableCommands[] = {
  {"getboolallnames", GETBOOLALLNAMES, 0}, // return a copy of all boolean tc names
  {"getfloatallnames", GETFLOATALLNAMES, 0}, // return a copy of all float tc names
  {"getnumtunables", GETNUMTUNABLES, 0},           // size of global tunable lists
  {"getnumbooltunables", GETNUMBOOLTUNABLES, 0},   // size of global bool tunable list
  {"getnumfloattunables", GETNUMFLOATTUNABLES, 0}, // size of global float tunable list
  {"getdescription", GETDESCRIPTION, 1},   // string (name) to string (description)
  {"getvaluebyname", GETVALUEBYNAME, 1},   // string (name) to value
  {"setvaluebyname", SETVALUEBYNAME, 2},   // string (name) x value to NULL
  {"gettypebyname", GETTYPEBYNAME, 1},     // string (name) to tc type (bool, float)
  {"getusebyname", GETUSEBYNAME, 1},       // string (name) to tunableUse (developerConstant vs. userConstant)
  {"getfloatrangebyname", GETFLOATRANGEBYNAME, 1}, // string (name) to float range

  {NULL, CMDERROR, 0}
};

int findCommand(Tcl_Interp *interp, int argc, char **argv) {
   ostrstream resstr;

   if (argc==0) {
      resstr << "USAGE: uimpd tclTunable <option> args\n" << ends;
      SetInterpResult(interp, resstr);
      return CMDERROR;
   }

   for (cmdTabEntry *cmd = TclTunableCommands; cmd!=NULL; cmd++) {
      if (0==strcmp(argv[0], cmd->cmdName)) {
         // we have a match
         if (argc-1 == cmd->numArgs)
            return cmd->index;

         // wrong # args
         resstr << argv[0] << ": wrong # args ("
             << argc-1 << "); should be "
             << cmd->numArgs << ".\n" << ends;
         SetInterpResult(interp, resstr);
         return CMDERROR;
      }
   }

   // could not find command
   resstr << "TclTunable: unknown option (" << argv[0] << ")\n" << ends;
   SetInterpResult(interp, resstr);
   return CMDERROR;
}



Tcl_Obj*
getBoolAllNames(Tcl_Interp* interp)
{
    unsigned nBoolTunables = tunableConstantRegistry::numBoolTunables();
    vector<tunableBooleanConstant> allBoolConstants = tunableConstantRegistry::getAllBoolTunableConstants();
    assert(allBoolConstants.size() == nBoolTunables);
    unsigned int i;

    Tcl_Obj* resultObj = Tcl_NewListObj(0, NULL);
    for( i = 0; i < nBoolTunables; i++ )
    {
        const char* strName = allBoolConstants[i].getName().string_of();
        Tcl_Obj* theName = Tcl_NewStringObj( strName, -1 );
        int appendRet = Tcl_ListObjAppendElement(interp, resultObj, theName );
        if( appendRet != TCL_OK )
        {
            // the error is already reflected in the interpreter's result
            break;
        }
    }
    return resultObj;
}


Tcl_Obj*
getFloatAllNames(Tcl_Interp* interp)
{
    unsigned nFloatTunables = tunableConstantRegistry::numFloatTunables();
    vector<tunableFloatConstant> allFloatConstants = tunableConstantRegistry::getAllFloatTunableConstants();
    assert(allFloatConstants.size() == nFloatTunables);
    unsigned int i;

    Tcl_Obj* resultObj = Tcl_NewListObj(0, NULL);
    for( i = 0; i < nFloatTunables; i++ )
    {
        Tcl_Obj* theName = Tcl_NewStringObj( allFloatConstants[i].getName().string_of(), -1 );
        int appendRet = Tcl_ListObjAppendElement(interp, resultObj, theName );
        if( appendRet != TCL_OK )
        {
            // the error is already reflected in the interpreter's result
            break;
        }
    }
    return resultObj;
}


int TclTunableCommand(ClientData, Tcl_Interp *interp,
		      int argc, char **argv) {
   // This is the entrypoint for the TclTunable command.
   // i.e. once installed into tcl, a tcl code call to "TclTunable" enters here...
   ostrstream resstr;
   bool resultSet = false;


   int commandIndex = findCommand(interp, argc-1, argv+1);
   if (commandIndex == CMDERROR) {
      resstr << "uimpd tclTunable: could not parse" << ends;
      SetInterpResult(interp, resstr);
      return TCL_ERROR;
   }

   int ret = TCL_OK;
   switch (commandIndex) {
      case GETBOOLALLNAMES: {
        Tcl_SetObjResult( interp, getBoolAllNames(interp) );
        resultSet = true;
        break;
      }

      case GETFLOATALLNAMES: {
        Tcl_SetObjResult( interp, getFloatAllNames(interp) );
        resultSet = true;
        break;
      }

      case GETNUMTUNABLES:
         resstr << tunableConstantRegistry::numTunables() << ends;
         break;

      case GETNUMBOOLTUNABLES:
         resstr << tunableConstantRegistry::numBoolTunables() << ends;
         break;

      case GETNUMFLOATTUNABLES:
         resstr << tunableConstantRegistry::numFloatTunables() << ends;
         break;

      case GETDESCRIPTION: {
         // string (name) --> description (if no description, we substitute the name)
 	     if (!tunableConstantRegistry::existsTunableConstant(argv[2])) {
                resstr << "tclTunable getdescription: unknown tunable "
                    << argv[2] << "\n" << ends;
                ret = TCL_ERROR;
                break;
	     }

         tunableConstantBase tcb = tunableConstantRegistry::getGenericTunableConstantByName(argv[2]);
	     resstr << const_cast<char*>((tcb.getDesc().string_of()==NULL) ? 
					 tcb.getName().string_of() : 
					 tcb.getDesc().string_of()) << ends;
         break;
      }

      case GETVALUEBYNAME:
         // string (name) --> string (value)
 	     if (!tunableConstantRegistry::existsTunableConstant(argv[2])) {
            resstr << "tclTunable getvaluebyname: unknown tunable "
                << argv[2] << "\n" << ends;
            ret = TCL_ERROR;
            break;
	     }

         if (tunableConstantRegistry::getTunableConstantType(argv[2]) == tunableBoolean) {
            tunableBooleanConstant tbc = tunableConstantRegistry::findBoolTunableConstant(argv[2]);
            if (tbc.getValue() == true)
                resstr << "1" << ends;
            else
                resstr << "0" << ends;
         }
         else {
            tunableFloatConstant tfc = tunableConstantRegistry::findFloatTunableConstant(argv[2]);
            resstr << tfc.getValue() << ends;
         }
         break;

      case SETVALUEBYNAME:
         // string (name) x string(value) --> NULL
 	     if (!tunableConstantRegistry::existsTunableConstant(argv[2])) {
            resstr << "tclTunable setvaluebyname: unknown tunable "
                << argv[2] << "\n" << ends;
            ret = TCL_ERROR;
            break;
	     }

         if (tunableConstantRegistry::getTunableConstantType(argv[2]) == tunableBoolean)
	       tunableConstantRegistry::setBoolTunableConstant(argv[2], (bool)atoi(argv[3]));
         else
	       tunableConstantRegistry::setFloatTunableConstant(argv[2], (float)atof(argv[3]));
         break;

      case GETTYPEBYNAME:
         // string (name) --> string (type)
 	     if (!tunableConstantRegistry::existsTunableConstant(argv[2])) {
                resstr << "tclTunable gettypebyname: unknown tunable "
                    << argv[2] << "\n" << ends;
                ret = TCL_ERROR;
                break;
	     }

         if (tunableConstantRegistry::getTunableConstantType(argv[2]) == tunableBoolean)
             resstr << "bool" << ends;
         else
             resstr << "float" << ends;
         break;

      case GETUSEBYNAME: {
             // string (name) --> string (use)
 	         if (!tunableConstantRegistry::existsTunableConstant(argv[2])) {
                    resstr << "tclTunable getusebyname: unknown tunable "
                        << argv[2] << "\n" << ends;
                    ret = TCL_ERROR;
                    break;
	         }

             tunableConstantBase tcb = tunableConstantRegistry::getGenericTunableConstantByName(argv[2]);
             if (tcb.getUse() == developerConstant)
                 resstr << "developer" << ends;
             else
                 resstr << "user" << ends;

         }
         break;

      case GETFLOATRANGEBYNAME: {
             // name --> string (float range)
             if (!tunableConstantRegistry::existsFloatTunableConstant(argv[2])) {
                resstr << "tclTunable getfloatrangebyname: unknown float tunable "
                    << argv[2] << "\n" << ends;
                ret = TCL_ERROR;
                break;
             }

             bool aflag;
	         aflag=(tunableFloat == tunableConstantRegistry::getTunableConstantType(argv[2]));
	         assert(aflag);
             tunableFloatConstant tfc = tunableConstantRegistry::findFloatTunableConstant(argv[2]);
             resstr << tfc.getMin() << " " << tfc.getMax() << ends;
         }
         break;

      default:
          resstr << "unrecognized tclTunable command '" << argv[1] << "'" << ends;
          ret = TCL_ERROR;
   }

   if( !resultSet )
   {
        SetInterpResult(interp, resstr);
   }
   return ret;
}

void InstallTunableTclCommand(Tcl_Interp *interp) {
   // Call this in your C++ program (presumably near the
   // beginning) to add the "TclTunable" command to your tcl scripts...

   Tcl_CreateCommand(interp,
		     "tclTunable", // name of command
		     TclTunableCommand,
		     NULL, // no client data
		     NULL // no delete command
                    );
}
