/*
 * Copyright (c) 1996 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */





/************************************************************************
 * queue.c: an implementation of a FIFO queue ADT.
 *
 * the implementation of the queue is array based. the head of the queue
 * points to the element at the top of the queue, the tail to the next
 * available space for a new element to add on to the queue. an empty
 * queue is indicated by one where the head and tail point to the same
 * element.
 *
 * queues can be created, destroyed, enqueued and dequeued. automatic
 * resizing takes place as needed. queues can also be scanned like
 * lists and elements in the middle deleted.
************************************************************************/





/************************************************************************
 * header files.
************************************************************************/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "kludges.h"
#include "queue.h"





/************************************************************************
 * symbolic constants.
************************************************************************/

static const unsigned QUEUE_SIZE = 2048;





/************************************************************************
 * macros.
************************************************************************/

#define CINCR(v,s) (((v)+1)%(s))
#define CDECR(v,s) (((v)+(s)-1)%(s))





/************************************************************************
 * Queue q_init(unsigned size)
 *
 * create a queue of the specified size. a pointer to this queue is
 * the return value. if a `size' of 0 is specified, a queue of size
 * `QUEUE_SIZE' is allocated. all queues will be resized by a factor
 * dependent on the initial size.
************************************************************************/

Queue
q_init(unsigned size)
{
    Queue    q;
    unsigned i;

    if (size == 0) {
        size = QUEUE_SIZE;
    }
    if (!(q = (Queue) malloc(sizeof(struct thr_queue)))) {
        perror("malloc");
        return 0;
    }
    if (!(q->items = (void **) calloc(size, sizeof(void *)))) {
        perror("malloc");
        return 0;
    }

    q->size = size;
    q->head = 0;
    q->tail = 0;

    for (i = 0; i < q->size; i++) {
        q->items[i] = 0;
    }

    return q;
}





/************************************************************************
 * void q_free(Queue q)
 *
 * free all elements in the queue and the queue itself. the queue `q'
 * will no longer exist.
************************************************************************/

void
q_free(Queue q)
{
    assert(q);

    (void) free(q->items);
    (void) free(q);

    return;
}





/************************************************************************
 * q_clear(Queue q)
 *
 * throw away all the elements in the queue `q' and empty out the queue.
 * this is a _very_ dangerous operation since there may be no way to
 * get at the elements currently in the queue.
************************************************************************/

void
q_clear(Queue q)
{
    assert(q);

    q->head = 0;
    q->tail = 0;

    return;
}





/************************************************************************
 * void q_copy(Queue dst, Queue src)
 *
 * copy elements from the source queue `src' to the destination queue
 * `dst'. the size of `dst' must be no smaller than that of `src'. all
 * existing elements of `dst' will be lost.
************************************************************************/

void
q_copy(Queue dst, Queue src)
{
    unsigned i;

    assert(src && dst && (src->size <= dst->size));

    dst->head = 0;
    dst->tail = 0;

    for (i = src->head; i != src->tail; i = CINCR(i, src->size)) {
        dst->items[dst->tail] = src->items[i];
        dst->tail++;
    }

    return;
}





/************************************************************************
 * void* q_dequeue(Queue q)
 *
 * get the data item from the head of the queue `q'. the data item in
 * the head is the return value. empty queues will return a null.
************************************************************************/

void*
q_dequeue(Queue q)
{
    void* data;

    assert(q);
    if (q_isempty(q)) {
        return 0;
    }

    data    = q->items[q->head];
    q->head = CINCR(q->head, q->size);

    return data;
}





/************************************************************************
 * void q_enqueue(Queue q, void* data)
 *
 * insert data item `data' into queue `q'. if the queue is full, it is
 * automatically expanded. if no space can be allocated, a fatal error
 * is raised.
************************************************************************/

void
q_enqueue(Queue q, void* data)
{
    Queue p;

    assert(q);

    if (CINCR(q->tail, q->size) == q->head) {
        p = q_init(2*q->size);
        assert(p);
        q_copy(p, q);

        (void) free(q->items);
        q->size  = p->size;
        q->head  = p->head;
        q->tail  = p->tail;
        q->items = p->items;
        (void) free(p);
    }

    q->items[q->tail] = data;
    q->tail           = CINCR(q->tail, q->size);

    return;
}





/************************************************************************
 * bool q_isempty(Queue q)
 *
 * return a boolean indicating if queue `q' is empty or not.
************************************************************************/

bool
q_isempty(Queue q)
{
    assert(q);
    return (q->head == q->tail);
}





/************************************************************************
 * void q_print(Queue q)
 *
 * print the items in the queue `q'.
************************************************************************/

void
q_print(Queue q)
{
    unsigned i;

    assert(q);

    printf("%p: ", q);
    if (q_isempty(q)) {
        printf("(empty)");
    }
    else {
        for (i = q->head; i != q->tail; i = CINCR(i, q->size)) {
            printf("%p ", q->items[i]);
        }
    }
    printf("\n");

    return;
}





/************************************************************************
 * qscan_t q_initscan(Queue q)
 *
 * start a scan on queue `q' by returning a scan handle for the first
 * queue item. no check is made to see if the queue is empty.
************************************************************************/

qscan_t
q_initscan(Queue q)
{
    assert(q);
    return (qscan_t) q->head;
}





/************************************************************************
 * qscan_t q_nextscan(Queue q, qscan_t qs)
 *
 * return the next scan handle for queue `q' given the current scan is
 * at `qs'. no checks are made for invalid `qs' values or for end of
 * scans.
************************************************************************/

qscan_t
q_nextscan(Queue q, qscan_t qs)
{
    assert(q);
    return (qscan_t) CINCR(qs, q->size);
}





/************************************************************************
 * bool q_isbadqs(Queue q, qscan_t qs)
 *
 * return true if `qs' is an invalid scan handle for queue `q'. handles
 * are invalid for an empty queue or if they lie outside the range of
 * the head and tail pointers of the queue.
************************************************************************/

bool
q_isbadqs(Queue q, qscan_t qs)
{
    if (q_isempty(q)) {
        return true;
    }

    if ((q->head < q->tail)
        && ((qs < q->head) || (qs >= q->tail))) {
        return true;
    }

    if ((q->head > q->tail)
        && (qs < q->head) && (qs >= q->tail)) {
        return true;
    }

    return false;
}





/************************************************************************
 * void* q_getitem(Queue q, qscan_t qs)
 *
 * return the element of queue `q' at scan handle `qs'. it is an error
 * to specify invalid `qs' values.
************************************************************************/

void*
q_getitem(Queue q, qscan_t qs)
{
    assert(q);

    if (q_isbadqs(q, qs)) {
        fprintf(stderr, "invalid scan element %u\n", qs);
        return 0;
    }

    return q->items[qs];
}





/************************************************************************
 * void* q_delitem(Queue q, qscan_t qs)
 *
 * remove item `qs' from queue `q'. if `qs' is an invalid handle, an
 * error is raised.
 *
 * for a valid `qs', the item in question is deleted from the queue
 * and the remaining items shifted appropriately.
************************************************************************/

void*
q_delitem(Queue q, qscan_t qs)
{
    unsigned i;
    void*    ptr;

    assert(q);

    if (q_isbadqs(q, qs)) {
        fprintf(stderr, "invalid scan element %u\n", qs);
        return 0;
    }

    if (qs == q->head)
       return q_dequeue(q);

    ptr = q->items[qs];
    for (i = qs; i != q->tail; i = CINCR(i, q->size)) {
        q->items[i] = q->items[CINCR(i, q->size)];
    }

    q->tail = CDECR(q->tail, q->size);
    return ptr;
}
