/*
 * Copyright (c) 1996-2000 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */
#include "threadP.h"
#include "waitset.h"


bool
WaitSet::Add( thread_t tid )
{
	bool ret = true;


	if( !Contains( tid ) )
	{
		ret = DoAdd( tid );
	}
	return ret;
}


void
WaitSet::Remove( thread_t tid )
{
	if( Contains( tid ) )
	{
		unsigned idx = DoRemove( tid );

		if( !thr_tab->IsWindowsMsgQ( tid ) )
		{
			// do a simple compaction over removed element
			// note that nObjects has already been decremented
			Shift( idx + 1, nObjects, idx );
		}
	}
}


void
WaitSet::Remove( const WaitSet& ws )
{
	WaitSetIterator	iter( ws );
	thread_t tid;

	while( iter.Next( &tid ) )
	{
		if( Contains( tid ) )
		{
			DoRemove( tid );
		}
	}
	Compact();
}



void
WaitSet::Clear( void )
{
	unsigned i;
	for( i = 0; i < WAITSET_MAX_OBJECTS; i++ )
	{
		waitInfo[i].Reset();
	}
	wmsgTid = THR_TID_UNSPEC;

	nWaitees = 0;
	nObjects = 0;
}


bool
WaitSet::Contains( thread_t tid ) const
{
	return ((Find( tid ) < WAITSET_MAX_OBJECTS) ||
			((tid != THR_TID_UNSPEC) && (tid == wmsgTid)) );
}


bool
WaitSet::DoAdd( thread_t tid )
{
	assert( !Contains( tid ) );

	if( thr_tab->IsWindowsMsgQ( tid ) )
	{
		wmsgTid = tid;
		nWaitees++;
	}
	else
	{
		// find free spot in the set
		if( nObjects != WAITSET_MAX_OBJECTS )
		{
			waitInfo[nObjects].Set( tid,
                thr_tab->GetIOBindable( tid )->GetWaiterDesc() );
			nObjects++;
			nWaitees++;
		}
		else
		{
			// we're out of space for the set
			return false;
		}
	}

	assert( nObjects <= WAITSET_MAX_OBJECTS);

	return true;
}


unsigned
WaitSet::DoRemove( thread_t tid )
{
	unsigned index;

	if( thr_tab->IsWindowsMsgQ( tid ) )
	{
		wmsgTid = THR_TID_UNSPEC;
		nWaitees--;
	}
	else
	{
		// find the given waitee in the set
		index = Find( tid );
		assert( index < WAITSET_MAX_OBJECTS);

		// remove item from the set
		waitInfo[index].Reset();
		nObjects--;
		nWaitees--;
	}

	// we assume that at some efficient point after calling us, 
	// the caller will compact the array
	assert( nWaitees >= 0 );
	assert( nObjects >= 0 );

	return index;
}


unsigned
WaitSet::FindNextValid( unsigned i ) const
{
	while( i < WAITSET_MAX_OBJECTS )
	{
		if( waitInfo[i].obj != INVALID_HANDLE_VALUE )
		{
			break;
		}
		i++;
	}
	return i;
}

unsigned
WaitSet::FindNextInvalid( unsigned i ) const
{
	while( i < WAITSET_MAX_OBJECTS )
	{
		if( waitInfo[i].obj == INVALID_HANDLE_VALUE )
		{
			break;
		}
		i++;
	}
	return i;
}



// this code assumes that the shift is
// downward - i.e., idx_newstart <= idx_start
void
WaitSet::Shift( unsigned idx_start, unsigned idx_end, unsigned idx_newstart )
{
	unsigned idx_delta = idx_start - idx_newstart;
	unsigned i;


	// shift elements downward
	for( i = idx_start; i <= idx_end; i++ )
	{
		unsigned newi = i - (idx_delta);

		waitInfo[newi] = waitInfo[i];
	}

	// "zero" old elements at end of array
	for( i = idx_end - idx_delta + 1; i <= idx_end; i++ )
	{
		waitInfo[i].Reset();
	}
}


void
WaitSet::Compact( void )
{
	unsigned istart;
	unsigned vstart;
	unsigned inext;


	// find the first invalid entry
	istart = FindNextInvalid( 0 );
	vstart = FindNextValid( istart );
	while( vstart < WAITSET_MAX_OBJECTS )
	{
		// find end of valid segment
		inext = FindNextInvalid( vstart );

		// shift valid segment to start at istart
		Shift( vstart, inext - 1, istart );

		// advance to next invalid entry
		istart = inext;
		vstart = FindNextValid( istart );
	}
}



unsigned
WaitSet::Find( thread_t tid ) const
{
	unsigned i;

	// we can't stop searching at nObjects, since we're
	// used sometimes on uncompacted arrays
	for( i = 0; i < WAITSET_MAX_OBJECTS; i++ )
	{
		if( waitInfo[i].tid == tid )
		{
			break;
		}
	}
	return i;
}


// TODO - incorporate last tid hint.
// hard to do with current implementation, since we don't know 
// what type of item it is, or where we are in the current cycle

WaitSet::WaitReturn
WaitSet::Wait( bool block, thread_t last_tid_hint, thread_t* tidp, tag_t* tagp ) const
{
	DWORD		waitRet;
	WaitReturn	ret;
	unsigned	i;


	// initialize the out-parameters
	assert( tidp != NULL );
	assert( tagp != NULL );
	*tidp = THR_TID_UNSPEC;
	*tagp = MSG_TAG_UNSPEC;


    for( i = 0; i < nObjects; i++ )
	{
		if( waitInfo[i].isSocket )
        {
            Socket* ps = thr_tab->GetSocket( waitInfo[i].tid );

            if( ps->HasData() )
            {
                // return this socket as ready for reading
			    *tidp = waitInfo[i].tid;
			    *tagp = thr_tab->GetBindable( *tidp )->GetTag();

                return WaitSet::WaitInput;
            }
		}
	}

	// set all sockets to provide notification for events of interest
	// (note that the WSAEventSelect call sets the associated
	// socket to non-blocking mode)
    HANDLE* waitObj = new HANDLE[nObjects];
    unsigned int nWaitObj = 0;
	for( i = 0; i < nObjects; i++ )
	{
		if( waitInfo[i].isSocket )
		{
			// set event for notification of events of interest
			Socket* ps = thr_tab->GetSocket( waitInfo[i].tid );
			assert( ps != NULL );
            ResetEvent( ps->hEvent );
			ps->SetEvents();
		}

        waitObj[nWaitObj] = waitInfo[i].obj;
        nWaitObj++;
	}

	// check for one of our handles to be signalled, blocking if desired
	if( wmsgTid != THR_TID_UNSPEC )
	{
		waitRet = MsgWaitForMultipleObjects( nWaitObj, (HANDLE*)waitObj,
												false,
												(block ? INFINITE : 0),
												QS_ALLINPUT );
	}
	else
	{
		waitRet = WaitForMultipleObjects( nWaitObj, (HANDLE*)waitObj,
												false,
												(block ? INFINITE : 0) );
	}

	// determine why we regained control
	if( (waitRet >= WAIT_OBJECT_0) && (waitRet <= WAIT_OBJECT_0 + nObjects) )
	{
		// we've got input
		ret = WaitSet::WaitInput;

		if( waitRet == (WAIT_OBJECT_0 + nObjects) )
		{
			// there's a message in the message queue
			*tidp = wmsgTid;
			*tagp = MSG_TAG_WMSG;
		}
		else
		{
			// one of the objects was signalled
			unsigned ndx = waitRet - WAIT_OBJECT_0;

            bool found = false;
            for( i = 0; i < nObjects; i++ )
            {
                if( waitInfo[i].obj == waitObj[ndx] )
                {
                    ndx = i;
                    found = true;
                    break;
                }
            }
            assert( found );

			*tidp = waitInfo[ndx].tid;
			*tagp = thr_tab->GetBindable( *tidp )->GetTag();

			// reset the event object associated with the signalled socket
			WSANETWORKEVENTS wsaNetEvents;
			if( WSAEnumNetworkEvents( thr_socket( *tidp ), waitInfo[ndx].obj, &wsaNetEvents ) == SOCKET_ERROR )
            {
		        thr_report_error( "WaitSet::Wait(): WSAEnumNetworkEvents returned %d\n", WSAGetLastError() );
		        abort();
            }

            Socket* ps = thr_tab->GetSocket( waitInfo[ndx].tid );
            ps->HasData( true );
        }
	}
	else if( waitRet == WAIT_TIMEOUT )
	{
		// the call timed out
		ret = WaitTimeout;
	}
	else if( waitRet == (DWORD)(-1) )
	{
		// the call failed
		ret = WaitError;
	}
	else
	{
		// we've seen a return value we don't handle -
		// possibly an abandoned mutex
		thr_report_error( "WaitSet::Wait() *** unrecognized MsgWaitForMultipleObjects return ***\n" );
		abort();
	}

	// reset all sockets to blocking mode
	for( i = 0; i < nObjects; i++ )
	{
		if( waitInfo[i].isSocket )
		{
			// reset to blocking mode
			Socket* ps = thr_tab->GetSocket( waitInfo[i].tid );
			assert( ps != NULL );
			ps->ResetEvents();
		}
	}

	return ret;
}



// note that our implementation uses the condition (index == ws.nObjects)
// to represent that we are done searching the objects, but that we
// still might return a valid tid if we're watching the message queue
bool
WaitSetIterator::Next( thread_t* tidp )
{
	bool bOK = ((index < ws.nObjects) ||
				((index == ws.nObjects) && (ws.wmsgTid != THR_TID_UNSPEC)) );

	if( bOK )
	{
		if( index < ws.nObjects  )
		{
			if( tidp != NULL )
			{
				*tidp = ws.waitInfo[index].tid;
			}
		}
		else
		{
			*tidp = ws.wmsgTid;
		}
		index++;
	}
	return bOK;
}


void
WaitSet::ItemInfo::Set( thread_t _tid, HANDLE h )
{
	tid = _tid;
	isSocket = thr_tab->IsSocket( tid );
    obj = h;
}






